try:
    import os
    from botiverse.models import TTS
    from playsound import playsound
    from botiverse.models import LSTMClassifier
    from botiverse.preprocessors import Vocalize, Wav2Vec,  Frequency
    from botiverse.bots.VoiceBot.utils import voice_input
except:
    pass

class SpeechClassifier():
    '''
    An interface for the speech classifier chatbot which classifies speech into one of a set of classes. Suitable when the
    number of classes is small and the words are easily pronounceable.
    '''
    def __init__(self, words, samplerate, duration, repr='wav2vec', machine='lstm', **kwargs):
        '''
        Initialize the dataset and its transformation for the speech classification process.
        
        :param words: A list of words which are the classes of the speech classifier.
        :type words: list
        :param samplerate: The sample rate of the audio files.
        :type samplerate: int
        :param duration: The duration of the audio files in milliseconds.
        :type duration: int
        :param repr: The representation to use for the audio files. Can be 'wav2vec', 'mfcc', 'spectrogram' or a custom representation
        :type repr: str or object
        :param machine: The machine learning model to use for classification. Can be 'lstm' or a custom model.
        :type machine: str or object
        '''
        self.words = words
        self.samplerate = samplerate
        self.duration = duration
        self.machine = machine
        if repr == 'wav2vec':
            self.transformer = Wav2Vec(samplerate, duration)
        elif repr == 'mfcc':
            self.transformer = Frequency(type='mfcc', sample_rate=samplerate, duration=duration, **kwargs)
        elif repr == 'spectrogram':
            self.transformer = Frequency(type='spectrogram', sample_rate=samplerate, duration=duration, **kwargs)
        elif type(repr) != str:
            self.transformer = repr
        else:
            raise ValueError(f"Invalid representation {repr}. Expected wav2vec, mfcc or spectrogram.")


    def generate_read_data(self, n=3, regenerate=False, force_download_noise=False, **kwargs):
        '''
        Generate synthetic audio data for the words specified during init and then corrupt it with noise and audio transformations.
        
        :param n: The number of audio files to generate for each word using audio transformations.
        :type n: int
        :param regenerate: Whether to regenerate the dataset even if it already exists.
        :type regenerate: bool
        :param force_download_noise: Whether to force download the noise dataset even if it already exists.
        :type force_download_noise: bool
        :param kwargs: Keyword arguments to be passed to the transformer (that puts audio in the chosen representation).
        
        :return: A tuple of the form (X, y) where X is a 3D numpy array representing the audio files and y is a 1D numpy array representing the classes of the audio files.
        :rtype: tuple of numpy.ndarray
        '''
        # if there is no dataset folder or if the regenerate flag is set, generate the dataset
        if regenerate or not os.path.exists('dataset'):
            V = Vocalize(self.words)
            Vocalize.corrupt_dataset(self.words, sample_rate=self.samplerate, force_download=force_download_noise)
        X, y = self.transformer.transform_list(self.words, n, **kwargs)
        return X, y

    def fit(self, X, y,  λ=0.001, α=0.01, hidden=128, patience=50, max_epochs=600, **kwargs):
        '''
        Train the speech classifier model.
        
        :param X: A 3D numpy array representing the audio files.
        :type X: numpy.ndarray
        :param y: A 1D numpy array representing the classes of the audio files.
        :type y: numpy.ndarray
        :param λ: The learning rate parameter.
        :type λ: float
        :param α: The regularization parameter.
        :type α: float
        :param hidden: The number of hidden units in the LSTM layer.
        :type hidden: int
        :param patience: The number of bad epochs to wait before early stopping.
        :type patience: int
        :param max_epochs: The maximum number of epochs to train for.
        :type max_epochs: int
        :param kwargs: Keyword arguments to be passed to the model's fit method.
        '''
        if self.machine == 'lstm':
            self.model = LSTMClassifier(X.shape[2], hidden, len(self.words))
            self.model.fit(X, y,  λ, α, max_epochs, patience, **kwargs)
        elif type(self.machine) != str:
            self.model = self.machine
            self.model.fit(X, y, **kwargs)
        else:
            raise ValueError(f"Invalid machine {self.machine}. Expected lstm or a custom model.")
    
    def save(self, path):
        '''
        Save the model to a file.
        
        :param path: The path to the file
        '''
        self.model.save(path+'.bot')
    
    def load(self, path, **kwargs):
        '''
        Load the model from a file.
        
        :param path: The path to the file
        :param kwargs: Keyword arguments to be passed to the model's load method.
        '''
        if self.machine == 'lstm':
            self.model = LSTMClassifier(**kwargs)
            self.model.load(path + '.bot')
        else:
            self.model = self.machine
            self.model.load(path + '.bot')
            
        
    
    def predict(self, path, index=False):
        '''
        Predict the class of the audio file at the given path.
        
        :param path: The path to the audio file to be classified.
        :type path: str
        :param index: Whether to return the index of the class or the class itself.
        :type index: bool
        
        :return: The class of the audio file at the given path.
        :rtype: str or int
        '''
        vec = self.transformer.transform(path, strict_duration=False)
        pred, prob = self.model.predict(vec)
        pred, prob = pred[0], prob[0]
        print("Probability of prediction: ", prob)
        return (self.words[pred], prob) if not index else (pred, prob)



try:
    import torchaudio
    import os
    from tqdm import tqdm
    import torch
    import librosa
    import numpy as np
    import matplotlib.pyplot as plt
    from torch.utils.data import Dataset
    from audiomentations import Compose, AddGaussianNoise, TimeStretch, PitchShift, Shift
except:
    pass
class Frequency():
    '''
    An interface for transforming audio files into frequency domain representations.
    '''
    def __init__(self, sample_rate=16000, duration=1, augment=None, type='spec', 
                 nmels=70, n_fft=720, hop_length=360, is_log=True, **kwargs):
        
        '''
        Initialize the frequency transformer. 
        
        :param sample_rate: The sample rate of the audio files.
        :type sample_rate: int
        :param duration: The duration of the audio files in seconds.
        :type duration: int
        :param augment: The audio augmentations to apply to the audio files.
        :type augment: audiomentations.Compose
        :param type: The type of frequency domain representation to use. Can be 'spec' for spectrogram or 'mfcc' for Mel-frequency cepstral coefficients.
        :type type: str
        :param nmels: The number of mel bins to use for the Mel-frequency cepstral coefficients.
        :type nmels: int
        :param n_fft: The number of samples to use for each frame of the spectrogram.
        :type n_fft: int
        :param hop_length: The number of samples to shift the window by between frames of the spectrogram.
        :type hop_length: int
        :param is_log: Whether to use a log scale for the spectrogram.
        :type is_log: bool
        :param kwargs: Keyword arguments to be passed to the frequency domain transformer.
        
        '''

        self.sample_rate = sample_rate
        self.duration = duration
        self.emb_dim = 768 * int(49 * self.duration)         # fact regarding wav2vec2-base-960h
        if augment is None:
            self.augment = Compose([
            TimeStretch(min_rate=0.8, max_rate=1.4, p=0.7),
            PitchShift(min_semitones=-4, max_semitones=1, p=0.8),
            Shift(min_fraction=-0.5, max_fraction=0.5, p=0.5),
            AddGaussianNoise(min_amplitude=0.001, max_amplitude=0.015, p=0.5),
        ])
        else:
            self.augment = augment
        
        if type == 'spectrogram':
            self.transform_func = torchaudio.transforms.MelSpectrogram(sample_rate=self.sample_rate, n_fft=n_fft, hop_length=hop_length, n_mels=nmels, **kwargs)
        elif type == 'mfcc':
            self.transform_func = torchaudio.transforms.MFCC(sample_rate=self.sample_rate, n_mfcc=nmels, melkwargs={'n_fft':n_fft, 'hop_length':hop_length}, **kwargs)
        else:   raise ValueError(f"Invalid type of transformation {type}. Expected spectrogram or mfcc")
        
        self.is_log = is_log

    def transform_list(self, words, n=4):
        '''
        Given a folder dataset with folders each containing audio files, this returns a table of spectra in the form of a numpy array X and a table of classes in the form of a numpy array y.
        
        :param words: A list of words which are the classes of the speech classifier.
        :type words: list
        :param n: The number of times to augment each audio file.
        :type n: int
        
        :return: A tuple of the form (X, y) where X is a 3D numpy array representing the audio files and y is a 1D numpy array representing the classes of the audio files.
        :rtype: tuple of numpy.ndarray
        '''        
        # may be needed in the future (draft)
        sounds_per_word = len(os.listdir(f"dataset/{words[0]}"))
        self.N = len(words) * sounds_per_word
        
        # main function
        x_data, y_data = [], []
        for word in tqdm(words):
            for i, file in enumerate(os.listdir(f"dataset/{word}")):
                if not file.endswith(".wav"): continue
                waveform, sr = torchaudio.load(f"dataset/{word}/{file}")
                waveform = torchaudio.transforms.Resample(sr, self.sample_rate)(waveform)

                if waveform.shape[0] == 2:
                    waveform = torch.mean(waveform, dim=0, keepdim=True)
                
                length = waveform.shape[1]
                sample_dur = int(self.duration * self.sample_rate)
                if length < sample_dur:
                    waveform = torch.cat((waveform, torch.zeros(1, sample_dur - length)), dim=1)
                elif length > sample_dur:
                    waveform = waveform[:, :sample_dur]     
                
                waveform = waveform.numpy()
                if n > 0:
                    for _ in range(n):
                        waveform = self.augment(waveform, sample_rate=self.sample_rate)
                        spectrum = self.transform_func(torch.from_numpy(waveform))
                        spectrum = librosa.power_to_db(spectrum[0]) if self.is_log else spectrum[0]
                        # transpose to get (time, freq) instead of (freq, time)
                        spectrum = spectrum.T
                        x_data.append(spectrum)
                        y_data.append(words.index(word))                
                else:
                    spectrum = self.transform_func(torch.from_numpy(waveform))
                    spectrum = librosa.power_to_db(spectrum[0]) if self.is_log else spectrum[0]
                    # transpose to get (time, freq) instead of (freq, time)
                    spectrum = spectrum.T
                    x_data.append(spectrum)
                    y_data.append(words.index(word))
                    
        x_data, y_data = np.array(x_data), np.array(y_data)
        
        return x_data, y_data


    
    def transform(self, path, strict_duration=False):
        '''
        Convert the audio file given in path into a frequency domain representation.
        
        :param path: The path to the audio file.
        :type path: str
        :param strict_duration: Whether to strictly use the duration specified during init or not. If True, then the audio file is padded with zeros if it is shorter than the duration and truncated if it is longer than the duration.
        :type strict_duration: bool
        
        :return: The frequency domain representation of the audio file as a 2D numpy array.
        :rtype: numpy.ndarray
        '''
        waveform, sr = torchaudio.load(path)
        waveform = torchaudio.transforms.Resample(sr, self.sample_rate)(waveform)
        if waveform.shape[0] == 2:  waveform = torch.mean(waveform, dim=0, keepdim=True)
        length = waveform.shape[1]
        if strict_duration:
            sample_dur = int(self.duration * self.sample_rate)
            if length < sample_dur:
                waveform = torch.cat((waveform, torch.zeros(1, sample_dur - length)), dim=1)
            elif length > sample_dur:
                waveform = waveform[:, :sample_dur]
            
        spectrum = self.transform_func(waveform)
        spectrum = librosa.power_to_db(spectrum[0].numpy()) if self.is_log else spectrum[0].numpy()
        spectrum = spectrum.T
        spectrum = spectrum[np.newaxis, ...]
        return spectrum
        


